import os

import pytest


@pytest.fixture(scope="session")
def s3cred() -> None:
    os.environ["AWS_REGION"] = "us-west-2"
    os.environ["AWS_ACCESS_KEY_ID"] = "AKIAX7EGEQ7FT6CLQGWH"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "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"


@pytest.fixture()
def s3_localstack(monkeypatch):
    monkeypatch.setenv("AWS_REGION", "us-east-2")
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "test")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "test")
    monkeypatch.setenv("AWS_ENDPOINT_URL", "http://localhost:4566")



import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import pyarrow

# TODO: implement this module in Rust land to avoid JSON serialization
# https://github.com/delta-io/delta-rs/issues/95


@dataclass
class DataType:
    """
    Base class of all Delta data types.
    """

    type: str

    def __str__(self) -> str:
        return f"DataType({self.type})"

    @classmethod
    def from_dict(cls, json_dict: Dict[str, Any]) -> "DataType":
        """
        Generate a DataType from a DataType in json format.

        :param json_dict: the data type in json format
        :return: the Delta DataType
        """
        type_class = json_dict["type"]
        if type_class == "map":
            key_type_dict = {"type": json_dict["keyType"]}
            value_type_dict = {"type": json_dict["valueType"]}
            value_contains_null = json_dict["valueContainsNull"]
            key_type = cls.from_dict(json_dict=key_type_dict)
            value_type = cls.from_dict(json_dict=value_type_dict)
            return MapType(
                key_type=key_type,
                value_type=value_type,
                value_contains_null=value_contains_null,
            )
        if type_class == "array":
            field = json_dict["elementType"]
            if isinstance(field, str):
                element_type = cls(field)
            else:
                element_type = cls.from_dict(json_dict=field)
            return ArrayType(
                element_type=element_type,
                contains_null=json_dict["containsNull"],
            )
        if type_class == "struct":
            fields = []
            for json_field in json_dict["fields"]:
                if isinstance(json_field["type"], str):
                    data_type = cls(json_field["type"])
                else:
                    data_type = cls.from_dict(json_field["type"])
                field = Field(
                    name=json_field["name"],
                    type=data_type,
                    nullable=json_field["nullable"],
                    metadata=json_field.get("metadata"),
                )
                fields.append(field)
            return StructType(fields=fields)

        return DataType(type_class)


@dataclass(init=False)
class MapType(DataType):
    """Concrete class for map data types."""

    key_type: DataType
    value_type: DataType
    value_contains_null: bool
    type: str

    def __init__(
        self, key_type: "DataType", value_type: "DataType", value_contains_null: bool
    ):
        super().__init__("map")
        self.key_type = key_type
        self.value_type = value_type
        self.value_contains_null = value_contains_null

    def __str__(self) -> str:
        return f"DataType(map<{self.key_type}, {self.value_type}, {self.value_contains_null}>)"


@dataclass(init=False)
class ArrayType(DataType):
    """Concrete class for array data types."""

    element_type: DataType
    contains_null: bool
    type: str

    def __init__(self, element_type: DataType, contains_null: bool):
        super().__init__("array")
        self.element_type = element_type
        self.contains_null = contains_null

    def __str__(self) -> str:
        return f"DataType(array<{self.element_type}> {self.contains_null})"


@dataclass(init=False)
class StructType(DataType):
    """Concrete class for struct data types."""

    fields: List["Field"]
    type: str

    def __init__(self, fields: List["Field"]):
        super().__init__("struct")
        self.fields = fields

    def __str__(self) -> str:
        field_strs = [str(f) for f in self.fields]
        return f"DataType(struct<{', '.join(field_strs)}>)"


@dataclass
class Field:
    """Create a DeltaTable Field instance."""

    name: str
    type: DataType
    nullable: bool
    metadata: Optional[Dict[str, str]] = None

    def __str__(self) -> str:
        return f"Field({self.name}: {self.type} nullable({self.nullable}) metadata({self.metadata}))"


@dataclass
class Schema:
    """Create a DeltaTable Schema instance."""

    fields: List[Field]
    json_value: Dict[str, Any]

    def __str__(self) -> str:
        field_strs = [str(f) for f in self.fields]
        return f"Schema({', '.join(field_strs)})"

    def json(self) -> Dict[str, Any]:
        return self.json_value

    @classmethod
    def from_json(cls, json_data: str) -> "Schema":
        """
        Generate a DeltaTable Schema from a json format.

        :param json_data: the schema in json format
        :return: the DeltaTable schema
        """
        json_value = json.loads(json_data)
        fields = []
        for json_field in json_value["fields"]:
            if isinstance(json_field["type"], str):
                data_type = DataType(json_field["type"])
            else:
                data_type = DataType.from_dict(json_field["type"])
            field = Field(
                name=json_field["name"],
                type=data_type,
                nullable=json_field["nullable"],
                metadata=json_field.get("metadata"),
            )
            fields.append(field)
        return cls(fields=fields, json_value=json_value)


def pyarrow_datatype_from_dict(json_dict: Dict[str, Any]) -> pyarrow.DataType:
    """
    Create a DataType in PyArrow format from a Schema json format.

    :param json_dict: the DataType in json format
    :return: the DataType in PyArrow format
    """
    type_class = json_dict["type"]["name"]
    if type_class == "dictionary":
        key_type = json_dict["dictionary"]["indexType"]
        value_type = json_dict["children"][0]
        key_type = pyarrow_datatype_from_dict(key_type)
        value_type = pyarrow_datatype_from_dict(value_type)
        return pyarrow.map_(key_type, value_type)
    elif "dictionary" in json_dict:
        key_type = {
            "name": "key",
            "type": json_dict["dictionary"]["indexType"],
            "nullable": json_dict["nullable"],
        }
        key = pyarrow_datatype_from_dict(key_type)
        if type_class == "list":
            value_type = {
                "name": "val",
                "type": json_dict["dictionary"]["indexType"],
                "nullable": json_dict["nullable"],
            }
            return pyarrow.map_(
                key,
                pyarrow.list_(
                    pyarrow.field(
                        "entries", pyarrow.struct([pyarrow_field_from_dict(value_type)])
                    )
                ),
            )
        value_type = {
            "name": "value",
            "type": json_dict["type"],
            "nullable": json_dict["nullable"],
        }
        return pyarrow.map_(key, pyarrow_datatype_from_dict(value_type))
    elif type_class == "list":
        field = json_dict["children"][0]
        element_type = pyarrow_datatype_from_dict(field)
        return pyarrow.list_(pyarrow.field("item", element_type))
    elif type_class == "struct":
        fields = [pyarrow_field_from_dict(field) for field in json_dict["children"]]
        return pyarrow.struct(fields)
    elif type_class == "int":
        return pyarrow.type_for_alias(f'{type_class}{json_dict["type"]["bitWidth"]}')
    elif type_class == "date":
        type_info = json_dict["type"]
        if type_info["unit"] == "DAY":
            return pyarrow.date32()
        else:
            return pyarrow.date64()
    elif type_class == "time":
        type_info = json_dict["type"]
        if type_info["unit"] == "MICROSECOND":
            unit = "us"
        elif type_info["unit"] == "NANOSECOND":
            unit = "ns"
        elif type_info["unit"] == "MILLISECOND":
            unit = "ms"
        else:
            unit = "s"
        return pyarrow.type_for_alias(f'{type_class}{type_info["bitWidth"]}[{unit}]')
    elif type_class == "timestamp":
        type_info = json_dict["type"]
        if "unit" in type_info:
            if type_info["unit"] == "MICROSECOND":
                unit = "us"
            elif type_info["unit"] == "NANOSECOND":
                unit = "ns"
            elif type_info["unit"] == "MILLISECOND":
                unit = "ms"
            elif type_info["unit"] == "SECOND":
                unit = "s"
        else:
            unit = "ns"
        return pyarrow.type_for_alias(f"{type_class}[{unit}]")
    elif type_class.startswith("decimal"):
        type_info = json_dict["type"]
        return pyarrow.decimal128(
            precision=type_info["precision"], scale=type_info["scale"]
        )
    elif type_class.startswith("floatingpoint"):
        type_info = json_dict["type"]
        if type_info["precision"] == "HALF":
            return pyarrow.float16()
        elif type_info["precision"] == "SINGLE":
            return pyarrow.float32()
        elif type_info["precision"] == "DOUBLE":
            return pyarrow.float64()
    else:
        return pyarrow.type_for_alias(type_class)


def pyarrow_field_from_dict(field: Dict[str, Any]) -> pyarrow.Field:
    """
    Create a Field in PyArrow format from a Field in json format.
    :param field: the field in json format
    :return: the Field in PyArrow format
    """
    return pyarrow.field(
        field["name"],
        pyarrow_datatype_from_dict(field),
        field["nullable"],
        field.get("metadata"),
    )


def pyarrow_schema_from_json(json_data: str) -> pyarrow.Schema:
    """
    Create a Schema in PyArrow format from a Schema in json format.

    :param json_data: the field in json format
    :return: the Schema in PyArrow format
    """
    schema_json = json.loads(json_data)
    arrow_fields = [pyarrow_field_from_dict(field) for field in schema_json["fields"]]
    return pyarrow.schema(arrow_fields)



import json
import uuid
from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from typing import Any, Dict, Iterable, Iterator, List, Mapping, Optional, Union

import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.fs as pa_fs
from pyarrow.lib import RecordBatchReader
from typing_extensions import Literal

from .deltalake import PyDeltaTableError
from .deltalake import write_new_deltalake as _write_new_deltalake
from .table import DeltaTable


class DeltaTableProtocolError(PyDeltaTableError):
    pass


@dataclass
class AddAction:
    path: str
    size: int
    partition_values: Mapping[str, Optional[str]]
    modification_time: int
    data_change: bool
    stats: str


def write_deltalake(
    table_or_uri: Union[str, DeltaTable],
    data: Union[
        pd.DataFrame,
        pa.Table,
        pa.RecordBatch,
        Iterable[pa.RecordBatch],
        RecordBatchReader,
    ],
    schema: Optional[pa.Schema] = None,
    partition_by: Optional[List[str]] = None,
    filesystem: Optional[pa_fs.FileSystem] = None,
    mode: Literal["error", "append", "overwrite", "ignore"] = "error",
    name: Optional[str] = None,
    description: Optional[str] = None,
    configuration: Optional[Mapping[str, Optional[str]]] = None,
) -> None:
    """Write to a Delta Lake table (Experimental)

    If the table does not already exist, it will be created.

    This function only supports protocol version 1 currently. If an attempting
    to write to an existing table with a higher min_writer_version, this
    function will throw DeltaTableProtocolError.

    Note that this function does NOT register this table in a data catalog.

    :param table_or_uri: URI of a table or a DeltaTable object.
    :param data: Data to write. If passing iterable, the schema must also be given.
    :param schema: Optional schema to write.
    :param partition_by: List of columns to partition the table by. Only required
        when creating a new table.
    :param filesystem: Optional filesystem to pass to PyArrow. If not provided will
        be inferred from uri.
    :param mode: How to handle existing data. Default is to error if table
        already exists. If 'append', will add new data. If 'overwrite', will
        replace table with new data. If 'ignore', will not write anything if
        table already exists.
    :param name: User-provided identifier for this table.
    :param description: User-provided description for this table.
    :param configuration: A map containing configuration options for the metadata action.
    """
    if isinstance(data, pd.DataFrame):
        data = pa.Table.from_pandas(data)

    if schema is None:
        if isinstance(data, RecordBatchReader):
            schema = data.schema
        elif isinstance(data, Iterable):
            raise ValueError("You must provide schema if data is Iterable")
        else:
            schema = data.schema

    if isinstance(table_or_uri, str):
        table = try_get_deltatable(table_or_uri)
        table_uri = table_or_uri
    else:
        table = table_or_uri
        table_uri = table_uri = table._table.table_uri()

    # TODO: Pass through filesystem once it is complete
    # if filesystem is None:
    #    filesystem = pa_fs.PyFileSystem(DeltaStorageHandler(table_uri))

    if table:  # already exists
        if mode == "error":
            raise AssertionError("DeltaTable already exists.")
        elif mode == "ignore":
            return

        current_version = table.version()

        if partition_by:
            assert partition_by == table.metadata().partition_columns

        if table.protocol().min_writer_version > 1:
            raise DeltaTableProtocolError(
                "This table's min_writer_version is "
                f"{table.protocol().min_writer_version}, "
                "but this method only supports version 1."
            )
    else:  # creating a new table
        current_version = -1

        # TODO: Don't allow writing to non-empty directory
        # Blocked on: Finish filesystem implementation in fs.py
        # assert len(filesystem.get_file_info(pa_fs.FileSelector(table_uri, allow_not_found=True))) == 0

    if partition_by:
        partition_schema = pa.schema([schema.field(name) for name in partition_by])
        partitioning = ds.partitioning(partition_schema, flavor="hive")
    else:
        partitioning = None

    add_actions: List[AddAction] = []

    def visitor(written_file: Any) -> None:
        partition_values = get_partitions_from_path(table_uri, written_file.path)
        stats = get_file_stats_from_metadata(written_file.metadata)

        add_actions.append(
            AddAction(
                written_file.path,
                written_file.metadata.serialized_size,
                partition_values,
                int(datetime.now().timestamp()),
                True,
                json.dumps(stats, cls=DeltaJSONEncoder),
            )
        )

    ds.write_dataset(
        data,
        base_dir=table_uri,
        basename_template=f"{current_version + 1}-{uuid.uuid4()}-{{i}}.parquet",
        format="parquet",
        partitioning=partitioning,
        # It will not accept a schema if using a RBR
        schema=schema if not isinstance(data, RecordBatchReader) else None,
        file_visitor=visitor,
        existing_data_behavior="overwrite_or_ignore",
    )

    if table is None:
        _write_new_deltalake(  # type: ignore[call-arg]
            table_uri,
            schema,
            add_actions,
            mode,
            partition_by or [],
            name,
            description,
            configuration,
        )
    else:
        table._table.create_write_transaction(
            add_actions,
            mode,
            partition_by or [],
        )


class DeltaJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, bytes):
            return obj.decode("unicode_escape")
        elif isinstance(obj, date):
            return obj.isoformat()
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, Decimal):
            return str(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def try_get_deltatable(table_uri: str) -> Optional[DeltaTable]:
    try:
        return DeltaTable(table_uri)
    except PyDeltaTableError as err:
        if "Not a Delta table" not in str(err):
            raise
        return None


def get_partitions_from_path(base_path: str, path: str) -> Dict[str, str]:
    path = path.split(base_path, maxsplit=1)[1]
    parts = path.split("/")
    parts.pop()  # remove filename
    out = {}
    for part in parts:
        if part == "":
            continue
        key, value = part.split("=", maxsplit=1)
        out[key] = value
    return out


def get_file_stats_from_metadata(
    metadata: Any,
) -> Dict[str, Union[int, Dict[str, Any]]]:
    stats = {
        "numRecords": metadata.num_rows,
        "minValues": {},
        "maxValues": {},
        "nullCount": {},
    }

    def iter_groups(metadata: Any) -> Iterator[Any]:
        for i in range(metadata.num_row_groups):
            yield metadata.row_group(i)

    for column_idx in range(metadata.num_columns):
        name = metadata.row_group(0).column(column_idx).path_in_schema
        # If stats missing, then we can't know aggregate stats
        if all(
            group.column(column_idx).is_stats_set for group in iter_groups(metadata)
        ):
            stats["nullCount"][name] = sum(
                group.column(column_idx).statistics.null_count
                for group in iter_groups(metadata)
            )

            # I assume for now this is based on data type, and thus is
            # consistent between groups
            if metadata.row_group(0).column(column_idx).statistics.has_min_max:
                # Min and Max are recorded in physical type, not logical type
                # https://stackoverflow.com/questions/66753485/decoding-parquet-min-max-statistics-for-decimal-type
                # TODO: Add logic to decode physical type for DATE, DECIMAL
                logical_type = (
                    metadata.row_group(0)
                    .column(column_idx)
                    .statistics.logical_type.type
                )
                #
                if logical_type not in ["STRING", "INT", "TIMESTAMP", "NONE"]:
                    continue
                # import pdb; pdb.set_trace()
                stats["minValues"][name] = min(
                    group.column(column_idx).statistics.min
                    for group in iter_groups(metadata)
                )
                stats["maxValues"][name] = max(
                    group.column(column_idx).statistics.max
                    for group in iter_groups(metadata)
                )
    return stats



import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pytest

from deltalake import DeltaTable
from deltalake.deltalake import DeltaStorageFsBackend
from deltalake.fs import DeltaStorageHandler


def test_normalize_path():
    backend = DeltaStorageFsBackend("")
    assert backend.normalize_path("s3://foo/bar") == "s3://foo/bar"
    assert backend.normalize_path("s3://foo/bar/") == "s3://foo/bar"
    assert backend.normalize_path("/foo/bar//") == "/foo/bar"


@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=5, method="thread")
def test_read_files(s3_localstack):
    table_path = "s3://deltars/simple"
    handler = DeltaStorageHandler(table_path)
    dt = DeltaTable(table_path)
    files = dt.file_uris()
    assert len(files) > 0
    for file in files:
        with handler.open_input_file(file) as f_:
            table = pq.read_table(f_)
            assert isinstance(table, pa.Table)
            assert table.shape > (0, 0)


@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=5, method="thread")
def test_read_simple_table_from_remote(s3_localstack):
    table_path = "s3://deltars/simple"
    dt = DeltaTable(table_path)
    assert dt.to_pandas().equals(pd.DataFrame({"id": [5, 7, 9]}))



from typing import Any, Callable, List

import pyarrow as pa

from deltalake.writer import AddAction

RawDeltaTable: Any
rust_core_version: Callable[[], str]
DeltaStorageFsBackend: Any

write_new_deltalake: Callable[[str, pa.Schema, List[AddAction], str, List[str]], None]

class PyDeltaTableError(BaseException): ...



import json
import os
import pathlib
import sys
from datetime import date, datetime, timedelta
from decimal import Decimal
from unittest.mock import Mock

import pyarrow as pa
import pyarrow.compute as pc
import pytest
from pandas.testing import assert_frame_equal
from pyarrow.lib import RecordBatchReader

from deltalake import DeltaTable, write_deltalake
from deltalake.table import ProtocolVersions
from deltalake.writer import DeltaTableProtocolError


def _is_old_glibc_version():
    if "CS_GNU_LIBC_VERSION" in os.confstr_names:
        version = os.confstr("CS_GNU_LIBC_VERSION").split(" ")[1]
        return version < "2.28"
    else:
        return False


if sys.platform == "win32":
    pytest.skip("Writer isn't yet supported on Windows", allow_module_level=True)

if _is_old_glibc_version():
    pytest.skip(
        "Writer isn't yet supported on Linux with glibc < 2.28", allow_module_level=True
    )


@pytest.fixture()
def sample_data():
    nrows = 5
    return pa.table(
        {
            "utf8": pa.array([str(x) for x in range(nrows)]),
            "int64": pa.array(list(range(nrows)), pa.int64()),
            "int32": pa.array(list(range(nrows)), pa.int32()),
            "int16": pa.array(list(range(nrows)), pa.int16()),
            "int8": pa.array(list(range(nrows)), pa.int8()),
            "float32": pa.array([float(x) for x in range(nrows)], pa.float32()),
            "float64": pa.array([float(x) for x in range(nrows)], pa.float64()),
            "bool": pa.array([x % 2 == 0 for x in range(nrows)]),
            "binary": pa.array([str(x).encode() for x in range(nrows)]),
            "decimal": pa.array([Decimal("10.000") + x for x in range(nrows)]),
            "date32": pa.array(
                [date(2022, 1, 1) + timedelta(days=x) for x in range(nrows)]
            ),
            "timestamp": pa.array(
                [datetime(2022, 1, 1) + timedelta(hours=x) for x in range(nrows)]
            ),
            "struct": pa.array([{"x": x, "y": str(x)} for x in range(nrows)]),
            "list": pa.array([list(range(x + 1)) for x in range(nrows)]),
            # NOTE: https://github.com/apache/arrow-rs/issues/477
            #'map': pa.array([[(str(y), y) for y in range(x)] for x in range(nrows)], pa.map_(pa.string(), pa.int64())),
        }
    )


@pytest.fixture()
def existing_table(tmp_path: pathlib.Path, sample_data: pa.Table):
    path = str(tmp_path)
    write_deltalake(path, sample_data)
    return DeltaTable(path)


@pytest.mark.skip(reason="Waiting on #570")
def test_handle_existing(tmp_path: pathlib.Path, sample_data: pa.Table):
    # if uri points to a non-empty directory that isn't a delta table, error
    tmp_path
    p = tmp_path / "hello.txt"
    p.write_text("hello")

    with pytest.raises(OSError) as exception:
        write_deltalake(str(tmp_path), sample_data, mode="overwrite")

    assert "directory is not empty" in str(exception)


def test_roundtrip_basic(tmp_path: pathlib.Path, sample_data: pa.Table):
    write_deltalake(str(tmp_path), sample_data)

    assert ("0" * 20 + ".json") in os.listdir(tmp_path / "_delta_log")

    delta_table = DeltaTable(str(tmp_path))
    assert delta_table.pyarrow_schema() == sample_data.schema

    table = delta_table.to_pyarrow_table()
    assert table == sample_data


def test_roundtrip_metadata(tmp_path: pathlib.Path, sample_data: pa.Table):
    write_deltalake(
        str(tmp_path),
        sample_data,
        name="test_name",
        description="test_desc",
        configuration={"configTest": "foobar"},
    )

    delta_table = DeltaTable(str(tmp_path))

    metadata = delta_table.metadata()

    assert metadata.name == "test_name"
    assert metadata.description == "test_desc"
    assert metadata.configuration == {"configTest": "foobar"}


@pytest.mark.parametrize(
    "column",
    [
        "utf8",
        "int64",
        "int32",
        "int16",
        "int8",
        "float32",
        "float64",
        "bool",
        "binary",
        "date32",
    ],
)
def test_roundtrip_partitioned(
    tmp_path: pathlib.Path, sample_data: pa.Table, column: str
):
    write_deltalake(str(tmp_path), sample_data, partition_by=[column])

    delta_table = DeltaTable(str(tmp_path))
    assert delta_table.pyarrow_schema() == sample_data.schema

    table = delta_table.to_pyarrow_table()
    table = table.take(pc.sort_indices(table["int64"]))
    assert table == sample_data


def test_roundtrip_multi_partitioned(tmp_path: pathlib.Path, sample_data: pa.Table):
    write_deltalake(str(tmp_path), sample_data, partition_by=["int32", "bool"])

    delta_table = DeltaTable(str(tmp_path))
    assert delta_table.pyarrow_schema() == sample_data.schema

    table = delta_table.to_pyarrow_table()
    table = table.take(pc.sort_indices(table["int64"]))
    assert table == sample_data


def test_write_modes(tmp_path: pathlib.Path, sample_data: pa.Table):
    path = str(tmp_path)

    write_deltalake(path, sample_data)
    assert DeltaTable(path).to_pyarrow_table() == sample_data

    with pytest.raises(AssertionError):
        write_deltalake(path, sample_data, mode="error")

    write_deltalake(path, sample_data, mode="ignore")
    assert ("0" * 19 + "1.json") not in os.listdir(tmp_path / "_delta_log")

    write_deltalake(path, sample_data, mode="append")
    expected = pa.concat_tables([sample_data, sample_data])
    assert DeltaTable(path).to_pyarrow_table() == expected

    write_deltalake(path, sample_data, mode="overwrite")
    assert DeltaTable(path).to_pyarrow_table() == sample_data


def test_writer_with_table(existing_table: DeltaTable, sample_data: pa.Table):
    write_deltalake(existing_table, sample_data, mode="overwrite")
    existing_table.update_incremental()
    assert existing_table.to_pyarrow_table() == sample_data


def test_fails_wrong_partitioning(existing_table: DeltaTable, sample_data: pa.Table):
    with pytest.raises(AssertionError):
        write_deltalake(
            existing_table, sample_data, mode="append", partition_by="int32"
        )


def test_write_pandas(tmp_path: pathlib.Path, sample_data: pa.Table):
    # When timestamp is converted to Pandas, it gets casted to ns resolution,
    # but Delta Lake schemas only support us resolution.
    sample_pandas = sample_data.to_pandas().drop(["timestamp"], axis=1)
    write_deltalake(str(tmp_path), sample_pandas)

    delta_table = DeltaTable(str(tmp_path))
    df = delta_table.to_pandas()
    assert_frame_equal(df, sample_pandas)


def test_write_iterator(
    tmp_path: pathlib.Path, existing_table: DeltaTable, sample_data: pa.Table
):
    batches = existing_table.to_pyarrow_dataset().to_batches()
    with pytest.raises(ValueError):
        write_deltalake(str(tmp_path), batches, mode="overwrite")

    write_deltalake(str(tmp_path), batches, schema=sample_data.schema, mode="overwrite")
    assert DeltaTable(str(tmp_path)).to_pyarrow_table() == sample_data


def test_write_recordbatchreader(
    tmp_path: pathlib.Path, existing_table: DeltaTable, sample_data: pa.Table
):
    batches = existing_table.to_pyarrow_dataset().to_batches()
    reader = RecordBatchReader.from_batches(sample_data.schema, batches)

    write_deltalake(str(tmp_path), reader, mode="overwrite")
    assert DeltaTable(str(tmp_path)).to_pyarrow_table() == sample_data


def test_writer_partitioning(tmp_path: pathlib.Path):
    test_strings = ["a=b", "hello world", "hello%20world"]
    data = pa.table(
        {"p": pa.array(test_strings), "x": pa.array(range(len(test_strings)))}
    )

    write_deltalake(str(tmp_path), data)

    assert DeltaTable(str(tmp_path)).to_pyarrow_table() == data


def get_stats(table: DeltaTable):
    log_path = table._table.table_uri() + "/_delta_log/" + ("0" * 20 + ".json")

    # Should only have single add entry
    for line in open(log_path, "r").readlines():
        log_entry = json.loads(line)

        if "add" in log_entry:
            return json.loads(log_entry["add"]["stats"])
    else:
        raise AssertionError("No add action found!")


def test_writer_stats(existing_table: DeltaTable, sample_data: pa.Table):
    stats = get_stats(existing_table)

    assert stats["numRecords"] == sample_data.num_rows

    assert all(null_count == 0 for null_count in stats["nullCount"].values())

    expected_mins = {
        "utf8": "0",
        "int64": 0,
        "int32": 0,
        "int16": 0,
        "int8": 0,
        "float32": 0.0,
        "float64": 0.0,
        "bool": False,
        "binary": "0",
        # TODO: Writer needs special decoding for decimal and date32.
        #'decimal': '10.000',
        # "date32": '2022-01-01',
        "timestamp": "2022-01-01T00:00:00",
        "struct.x": 0,
        "struct.y": "0",
        "list.list.item": 0,
    }
    assert stats["minValues"] == expected_mins

    expected_maxs = {
        "utf8": "4",
        "int64": 4,
        "int32": 4,
        "int16": 4,
        "int8": 4,
        "float32": 4.0,
        "float64": 4.0,
        "bool": True,
        "binary": "4",
        #'decimal': '40.000',
        # "date32": '2022-01-04',
        "timestamp": "2022-01-01T04:00:00",
        "struct.x": 4,
        "struct.y": "4",
        "list.list.item": 4,
    }
    assert stats["maxValues"] == expected_maxs


def test_writer_null_stats(tmp_path: pathlib.Path):
    data = pa.table(
        {
            "int32": pa.array([1, None, 2, None], pa.int32()),
            "float64": pa.array([1.0, None, None, None], pa.float64()),
            "str": pa.array([None] * 4, pa.string()),
        }
    )
    path = str(tmp_path)
    write_deltalake(path, data)

    table = DeltaTable(path)
    stats = get_stats(table)

    expected_nulls = {"int32": 2, "float64": 3, "str": 4}
    assert stats["nullCount"] == expected_nulls


def test_writer_fails_on_protocol(existing_table: DeltaTable, sample_data: pa.Table):
    existing_table.protocol = Mock(return_value=ProtocolVersions(1, 2))
    with pytest.raises(DeltaTableProtocolError):
        write_deltalake(existing_table, sample_data, mode="overwrite")
